/*
 *                          Yeppp! library header
 *                  This file is auto-generated by Peach-Py,
 *       Portable Efficient Assembly Code-generator in Higher-level Python,
 *                   part of the Yeppp! library infrastrure
 *
 * This file is part of Yeppp! library and licensed under 2-clause BSD license.
 *
 * Copyright (C) 2010-2012 Marat Dukhan
 * Copyright (C) 2012-2013 Georgia Institute of Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are those
 * of the authors and should not be interpreted as representing official policies,
 * either expressed or implied, of the Yeppp! Project.
 */

#pragma once

#include <yepPredefines.h>
#include <yepTypes.h>

#ifdef __cplusplus
	extern "C" {
#endif

/** @defgroup yepMath yepMath: vector mathematical functions. */

/** @name	Natural Logarithm */
/**@{*/
/**
 * @ingroup	yepMath
 * @brief	Computes vector logarithm of double-precision (64-bit) floating-point elements in the input array, producing an array of double-precision (64-bit) floating-point elements in the output array.
 * @param[in]	numberPointer	Pointer the input array of double-precision (64-bit) floating-point elements to compute logarithms on.
 * @param[out]	logNumberPointer	Pointer the output array of double-precision (64-bit) floating-point elements to store the logarithm values.
 * @param[in]	length	The length of the arrays pointed by @a numberPointer and @a logNumberPointer.
 * @retval	#YepStatusOk	The computations finished successfully.
 * @retval	#YepStatusNullPointer	One of the @a numberPointer or @a logNumberPointer arguments is null.
 * @retval	#YepStatusMisalignedPointer	One of the @a numberPointer or @a logNumberPointer arguments is not properly aligned.
 * @par	Optimized implementations
 * 		<table>
 * 			<tr><th>Architecture</th><th>Target microarchitecture</th><th>Required instruction extensions</th></tr>
 * 			<tr><td>x86-64</td><td>Intel Nehalem</td><td>SSE, SSE4.2, SSE4.1, SSE2</td></tr>
 * 		</table>
 */
YEP_PUBLIC_SYMBOL enum YepStatus YEPABI yepMath_Log_V64f_V64f(const Yep64f* YEP_RESTRICT numberPointer, Yep64f* YEP_RESTRICT logNumberPointer, YepSize length);
/**@}*/

#ifdef __cplusplus
	} // extern "C"
#endif